#!/usr/bin/env python
"""
    Author: Mark Bokil
    email: mark@markbokil.org
    website: http://markbokil.org
    License: GPL 2
    Description: This is a volume control adesklet to set your volume. 
    Functions include mute, increase, decrease, display volume.
    Dependencies: adesklets
    Images: created using the Gimp 2

"""

import adesklets,string, os, sys, re, threading
import ossaudiodev

from os.path import join, dirname
from os import system, getenv


class Config(adesklets.ConfigFile):
    """
    This is the volume.py desklet configuration file.  
    """
    cfg_default = {
        'color'     : '#000000',
        'font'      : 'VeraBd',
        'font_size' : 8,
        'theme': 'default',
    }
        

class Events(adesklets.Events_handler):
    sizeX = 220
    sizeY = 52
    
    def __init__(self, basedir):
        """
        Constructor.  Set up the bare essentials.
        """
        if len(basedir) == 0:
            self.basedir = '.'
        else:
            self.basedir = basedir
        self.id = adesklets.get_id()
        self.editor = None
        self.config = None
        self.red    = 0
        self.green  = 0
        self.blue   = 0
        self.font = 'VeraBd'
        self.font_size = 8
        self.theme = 'default'
        self.volume = None
        self.mixer = None
        self.mute = False
        self.buffer = None
        self.clearRollovers()
        self.mousedown = False
        self.mouseover = False
        self.state = 0
        self.errorMixer = 'An error occured. Python sound mixer library ossaudiodev is missing.'
        adesklets.Events_handler.__init__(self)


    def clearRollovers(self):
        self.stateUp = False
        self.stateDown = False
        self.stateBar1 = False
        self.stateBar2 = False
        self.stateBar3 = False
        self.stateBar4 = False
        self.stateBar5 = False
        self.stateBar6 = False


    def loadImages(self):
        self.bg = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bg.png'))
        self.bgOff = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bg-off.png'))
        self.bar1 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar1.png'))
        self.bar2 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar2.png'))
        self.bar3 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar3.png'))
        self.bar4 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar4.png'))
        self.bar5 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar5.png'))
        self.bar6 = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar6.png'))
        self.bar1Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar1-over.png'))
        self.bar2Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar2-over.png'))
        self.bar3Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar3-over.png'))
        self.bar4Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar4-over.png'))
        self.bar5Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar5-over.png'))
        self.bar6Over = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/bar6-over.png'))
        self.btnDown = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/volume-down-over.png'))
        self.btnUp = adesklets.load_image(join(self.basedir, 'images/' + self.theme + '/volume-up-over.png'))


    def __del__(self):
        # free up image objects for GC, close mixer 
        adesklets.images_reset_all()
        self.mixer.close()
        adesklets.Events_handler.__del__(self)
       

    def ready(self):
        # initialize config
        self.id = adesklets.get_id()
        self._get_config()

        # open sound mixer control
        self.mixer = ossaudiodev.openmixer()

        # load images
        self.loadImages()

        # Set up the window
        self.buffer = adesklets.create_image(self.sizeX, self.sizeY)
        self.volume = self.get_volume()
        if self.volume == 0: self.mute = True
        adesklets.window_resize(self.sizeX, self.sizeY)
        adesklets.window_set_transparency(True)
        adesklets.window_show()

        # copy background to image buffer and display it
        adesklets.context_set_image(self.buffer)
        adesklets.context_set_blend(True)
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()

        # build context menu
        adesklets.menu_add_separator()
        adesklets.menu_add_item('README')
        adesklets.menu_add_submenu('Theme')
        adesklets.menu_add_item('Default')
        adesklets.menu_add_item('Graphite')
        adesklets.menu_add_item('Brushed')
        adesklets.menu_add_item('Simple')
        adesklets.menu_add_item('Dark')
        adesklets.menu_add_item('Blue')
        adesklets.menu_end_submenu()
        
        # editor setup
        editor=getenv('EDITOR')
        if editor:
            self.editor = editor
        else :
            self.editor = 'gedit'


    # copy buffer to foreground image updating UI
    def display(self) :
        adesklets.context_set_image(0)
        adesklets.context_set_blend(False)
        adesklets.blend_image_onto_image(self.buffer,1,0,0,self.sizeX,self.sizeY,0,0,self.sizeX,self.sizeY)
        adesklets.context_set_blend(True)


    # copy the background image to the buffer
    def blendBackground(self) :
        adesklets.free_image(self.buffer)
        self.buffer = adesklets.create_image(self.sizeX, self.sizeY)
        adesklets.context_set_image(self.buffer)
        if self.mute:
            adesklets.blend_image_onto_image(self.bgOff,1,0,0,self.sizeX,self.sizeY,0,0,self.sizeX,self.sizeY)
        else:
            adesklets.blend_image_onto_image(self.bg,1,0,0,self.sizeX,self.sizeY,0,0,self.sizeX,self.sizeY)
  

    def blendVolumeBar(self, image,x,start=51) :
        adesklets.context_set_image(self.buffer)
        adesklets.blend_image_onto_image(image,1,0,0,x,43,start,0,x,43)
 

    # copy the volume button images to the buffer
    def blendVolumeBtn(self, image,x,start=129) :
        adesklets.context_set_image(self.buffer)
        self.updateVolumeBars(self.volume)
        adesklets.blend_image_onto_image(image,1,0,0,x,43,start,0,x,43)


    def drawVolumeTxt(self, volume) :
        if self.mute == True: volume = 0;
        if volume > 100:
            volume = 100
        if volume < 0:
            volume = 0
        adesklets.context_set_image(self.buffer)
        adesklets.context_set_blend(False)
        font = '%s/%s' % (self.font, self.font_size)
        adesklets.load_font(font)
        adesklets.context_set_color(self.red,self.green,self.blue,200)

        # Draw text strings
        adesklets.context_set_font(0)
        volume = str(volume) + '%'
        adesklets.text_draw(169,18,volume)
        adesklets.free_font(0)


    # Set the current system volume.
    def set_volume(self, val) :
        if val > 100:
            val = 100
        if val < 0:
            val = 0

        try :
            self.mixer.set(ossaudiodev.SOUND_MIXER_VOLUME, (val,val))
        except IOError :
            print self.errorMixer


    # Get the current system volume.
    def get_volume(self) :
        try :
            s = self.mixer.get(ossaudiodev.SOUND_MIXER_VOLUME)
            return int(s[0])
        except IOError: 
            print self.errorMixer


    # mouse up/down state is tracked to avoid conflict between motion_notify and button_release events 
    # both were trying to use image blending at the same time creating errors, crashes in adesklets
    def setMouseState(self):
        self.mousedown = False
    
    def button_release(self, delayed, x, y, button):
        if button == 1:
            mouseTimer = threading.Timer(0.1, self.setMouseState)
            mouseTimer.start()
            try:
                self.state.cancel()
                self.state = 0
            except AttributeError:
                pass


    def button_press(self, delayed, x, y, button):
        if button==1:
             self.mousedown = True
             # mouseclick out of bounds
             if x >=160 :
                 return
             # volume on/off toggle area
             if x <= 53:
                  if self.mute == False:
                      self.mute = True
                      self.set_volume(0) 
                      self.blendBackground()
                      self.updateVolumeBars(self.volume)    
                      self.drawVolumeTxt(0)
                      self.display()

                  else:
                      self.mute = False
                      self.blendBackground()
                      self.set_volume(self.volume)
                      self.updateVolumeBars(self.volume)
                      self.drawVolumeTxt(self.volume)
                      self.display()
                  return

             # volume increase area
             if x >=130 and x <=145:
                 self.mute = False
                 if self.volume == 100: return
                 self.volume = self.volume + 1
                 self.set_volume(self.volume)
                 self.blendBackground()
                 self.updateVolumeBars(self.volume)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 if self.state != 0: return
                 self.state = threading.Timer(0.05, self.volumeIncrease)
                 self.state.start() 
                 return

             # volume decrease
             if x >=146 and x <=159:
                 if self.volume == 0: return
                 self.volume = self.volume - 1
                 self.set_volume(self.volume)
                 self.blendBackground()
                 self.updateVolumeBars(self.volume)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 if self.state != 0: return
                 self.state = threading.Timer(0.05, self.volumeDecrease)
                 self.state.start() 
                 return

             # mouseclick on volume slider 
             volume = int((x-63) * (x / 70.0))
             if volume > 100:
                 volume = 100
             if volume < 0:
                 volume = 0
                          
             self.mute = False
             self.volume = volume
             self.state = 0
             self.set_volume(self.volume)
             self.blendBackground()
             self.updateVolumeBars(self.volume)
             self.drawVolumeTxt(self.volume)
             self.display()
            

    def updateVolumeBars(self, volume) :
        ''' 
        draw bars on desklet based on volume strength 
        '''
        if volume > 100:
            volume = 100
        if volume < 0:
            volume = 0

        if volume <= 4 :
             self.blendBackground()
             return
        if volume <= 15 : 
             self.blendVolumeBar(self.bar1, 19)
             return   
        if volume <= 30 :
             self.blendVolumeBar(self.bar2, 29)
             return 
        if volume <= 45 :
             self.blendVolumeBar(self.bar3, 40)
             return  
        if volume <= 60 :
             self.blendVolumeBar(self.bar4, 51)
             return
        if volume <= 85 :
             self.blendVolumeBar(self.bar5, 61)
             return 
        if volume <= 100 :
             self.blendVolumeBar(self.bar6, 72)
       

    def volumeIncrease(self):
        self.volume = self.volume + 2
        if self.volume > 100: 
            self.volume = 100
        self.set_volume(self.volume) 
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()
        if self.state != 0:
            self.state = threading.Timer(0.1, self.volumeIncrease)
            self.state.start() 
     

    def volumeDecrease(self):
        self.volume = self.volume - 2
        if self.volume < 0: 
            self.volume = 0
        self.set_volume(self.volume) 
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()
        if self.state != 0:
            self.state = threading.Timer(0.1, self.volumeDecrease)
            self.state.start()        


    def quit(self):
        print 'Quitting...'
        

    def alarm(self):
        """
        When the alarm goes off, this is what we do.
        """
        # prevent alarm thread and mousedown thread from colliding
        if self.mouseover == True: return 1
        self.block()
        if self.mute == False: self.volume = self.get_volume()
        self.clearRollovers()
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()
        self.unblock()
        return 1


    # update GUI when leaving desklet
    def leave_notify(self, delayed, x, y):
        self.mouseover = False
        self.clearRollovers()
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()


    # update GUI when entering desklet
    def enter_notify(self, delayed, x, y):
        self.mouseover = True
        self.clearRollovers()
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()


    def motion_notify(self, delayed, x, y):
        ''' mouse position event processing '''
        if self.mousedown == True: return
        
        if y >= 9 and y <= 36:
            # blendVolumeBtn(image, width, x position)
            # note: the clearRollovers code prevents image re-blending redundancy to improve redraw performance
            # up /down buttons area
            if x >= 129 and x < 145: 
                 if self.stateUp == True : return
                 self.clearRollovers()
                 self.stateUp = True
                 self.blendBackground()
                 self.blendVolumeBtn(self.btnUp, 16,129)
                 self.drawVolumeTxt(self.volume)
                 self.display() 
                 self.stateUp = True
                 return   
            if x >= 145 and x <= 160: 
                 if self.stateDown == True : return
                 self.clearRollovers() 
                 self.stateDown = True
                 self.blendBackground()
                 self.blendVolumeBtn(self.btnDown, 16,145)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 self.btnStateDown = True
                 return   
            # volume bars area
            if x >= 60 and x <= 73:
                 if self.stateBar1 == True : return 
                 self.clearRollovers()
                 self.stateBar1 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar1Over, 19)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return   
            if x >=74 and x <= 83:
                 if self.stateBar2 == True : return 
                 self.clearRollovers()
                 self.stateBar2 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar2Over, 29)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return 
            if x >=84 and x <= 94:
                 if self.stateBar3 == True : return 
                 self.clearRollovers()
                 self.stateBar3 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar3Over, 40)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return 
            if x >=95 and x <= 105:
                 if self.stateBar4 == True : return 
                 self.clearRollovers()
                 self.stateBar4 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar4Over, 51)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return 
            if x >=106 and x <= 115:
                 if self.stateBar5 == True : return 
                 self.clearRollovers()
                 self.stateBar5 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar5Over, 61)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return 
            if x >=116 and x <= 126:
                 if self.stateBar6 == True : return 
                 self.clearRollovers()
                 self.stateBar6 = True
                 self.blendBackground()
                 self.blendVolumeBar(self.bar6Over, 72)
                 self.drawVolumeTxt(self.volume)
                 self.display()
                 return 

        # mouse position outside of volume buttons - update UI, clear rollover images
        if self.stateUp == True or self.stateDown == True or self.stateBar1 == True or  self.stateBar2 == True or self.stateBar3 == True or self.stateBar4 == True or self.stateBar5 == True or self.stateBar6 == True:
            self.clearRollovers()
            self.blendBackground()
            self.updateVolumeBars(self.volume)
            self.drawVolumeTxt(self.volume)
            self.display()
 

    #context menu
    def menu_fire(self, delayed, menu_id, item, something=0):
        if item == 'README' :
            system('xterm -e less %s/README &' % (self.basedir))
        elif item== 'Configure':
		    system('%s %s/config.txt &' % (self.editor, self.basedir ))
        elif item == 'Graphite':
            self.theme = 'graphite'
            self.switchTheme()
        elif item == 'Brushed':
            self.theme = 'brushed'
            self.switchTheme()
        elif item == 'Default':
            self.theme = 'default'
            self.switchTheme()
        elif item == 'Simple':
            self.theme = 'simple'
            self.switchTheme() 
        elif item == 'Blue':
            self.theme = 'blue'
            self.switchTheme()   
        elif item == 'Dark':
            self.theme = 'dark'
            self.switchTheme()   

    def switchTheme(self):
        self._set_config()
        adesklets.images_reset_all()
        self.loadImages()
        self.buffer = adesklets.create_image(self.sizeX, self.sizeY)
        self.blendBackground()
        self.updateVolumeBars(self.volume)
        self.drawVolumeTxt(self.volume)
        self.display()


    def _get_config(self):
        """
        Read configuration info from the configuration file.
        """
        if self.config is None:
            self.config = Config(adesklets.get_id(),
                    join(self.basedir,'config.txt'))
        config = self.config
        
        color = config.get('color')
        (self.red, self.green, self.blue) = self._parse_color(color)
        self.font = config.get('font')
        self.font_size = config.get('font_size')
        self.theme = config.get('theme')


    def _set_config(self):
        """
        Set configuration items to file
        """
        if self.config is None:
            self.config = Config(adesklets.get_id(),
                    join(self.basedir,'config.txt'))
        config = self.config
        color = "#FFFFFF"
        if self.theme == "graphite":
            config.__setitem__('color', color)
        else:
            color = "#000000"
            config.__setitem__('color', color)
       
        (self.red, self.green, self.blue) = self._parse_color(color)
        config.__setitem__('theme', self.theme)


    def _parse_color(self, color):
        """
        Parse an HTML-style hexadecimal color value into respective decimal RGB
        values.
        """
        def hex_color_to_int(color):
            return eval('0x%s' % color)
        color = color[1:] # remove "#" from start of color
        red = hex_color_to_int(color[:2])
        green = hex_color_to_int(color[2:4])
        blue = hex_color_to_int(color[4:])
        return (red, green, blue)


# pause the application waiting for alarm and user events
Events(dirname(__file__)).pause()
